from fastapi import Depends

from id_broker_common.kelvin import kelvin_session
from provisioning_plugin.models import NonEmptyStr, NoStarStr
from provisioning_plugin.utils import get_source_uid_for_authority, search_kelvin_obj
from ucsschool.apis.models import AuthenticatedUser
from ucsschool.apis.opa import OPAClient, opa_instance
from ucsschool.apis.utils import auth_manager
from ucsschool.kelvin.client import (
    School as KelvinSchool,
    SchoolClass as KelvinSchoolClass,
    SchoolClassResource,
    SchoolResource,
    Session,
    User as KelvinUser,
    UserResource,
)


async def check_for_authority_admin(
    school_authority: NonEmptyStr,
    policy_instance: OPAClient = Depends(opa_instance),
    user: AuthenticatedUser = Depends(auth_manager("provisioning", AuthenticatedUser)),
) -> bool:
    """
    Checks if the user is a school authority admin via OPA.
    """
    return await policy_instance.check_policy_true_or_raise(
        "provisioning_plugin/school_authority_admin",
        {"username": user.username, "school_authority": school_authority},
    )


async def get_school_class_by_name(
    name: NoStarStr,
    school_authority: NonEmptyStr,
    school: NoStarStr,
    session: Session = Depends(kelvin_session),
    _is_admin: bool = Depends(check_for_authority_admin),
) -> KelvinSchoolClass:
    return await search_kelvin_obj(
        school_authority,
        SchoolClassResource(session=session),
        name=name,
        school=f"{school_authority}-{school}",
    )


async def get_school_by_name(
    name: NonEmptyStr,
    school_authority: NonEmptyStr,
    session: Session = Depends(kelvin_session),
    _is_admin: bool = Depends(check_for_authority_admin),
) -> KelvinSchool:
    return await search_kelvin_obj(
        school_authority,
        SchoolResource(session=session),
        name=f"{school_authority}-{name}",
    )


async def get_user_by_id(
    user_id: NoStarStr,
    school_authority: NonEmptyStr,
    session: Session = Depends(kelvin_session),
    _is_admin: bool = Depends(check_for_authority_admin),
) -> KelvinUser:
    return await search_kelvin_obj(
        school_authority,
        UserResource(session=session),
        record_uid=user_id,
        source_uid=get_source_uid_for_authority(school_authority),
    )
