from typing import List, Set

from fastapi import APIRouter, Depends, HTTPException
from pydantic import ValidationError
from starlette import status
from starlette.responses import Response

from id_broker_common.kelvin import kelvin_session
from provisioning_plugin.dependencies import check_for_authority_admin, get_school_class_by_name
from provisioning_plugin.models import NonEmptyStr, NoStarStr, SchoolClass
from provisioning_plugin.utils import search_kelvin_obj, get_source_uid_for_authority
from ucsschool.kelvin.client import (
    InvalidRequest,
    NoObject,
    SchoolClass as KelvinSchoolClass,
    SchoolResource,
    Session,
    UserResource,
)

router = APIRouter(tags=["school_classes"])


async def member_ids_to_usernames(
    members: Set[str], school_authority: str, session: Session
) -> List[str]:
    """
    Takes a list of user_ids and returns a list of corresponding usernames.

    :param members: The list of user_ids to fetch the usernames for
    :param school_authority: The school authority to search in
    :param session: The Kelvin session to search with
    :return: The list of usernames
    :raises NoObject: If one of the members cannot be found
    """
    return [
        (
            await search_kelvin_obj(
                school_authority,
                UserResource(session=session),
                record_uid=user_id,
                source_uid=get_source_uid_for_authority(school_authority),
            )
        ).name
        for user_id in members
    ]


@router.get("/classes/{school}/{name}", response_model=SchoolClass)
async def get(
    name: NoStarStr,
    school_authority: NonEmptyStr,
    school: NoStarStr,
    school_class: KelvinSchoolClass = Depends(get_school_class_by_name),
    session: Session = Depends(kelvin_session),
) -> SchoolClass:
    # This try-except should be removed if Bug #53596 is fixed in Kelvin
    try:
        school = search_kelvin_obj(school_authority, SchoolResource(session=session), name=name)
    except HTTPException as exc:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"School class {name} in {school} could not be found.",
        ) from exc
    try:
        return SchoolClass.without_prefix(
            await SchoolClass.from_kelvin_school_class(
                school_class.as_dict(), UserResource(session=session)
            ),
            school_authority,
        )
    except (ValidationError, ValueError, NoObject) as exc:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=f"The school_class {name} in school {school} in school_authority "
            f"{school_authority} is malformed.",
        ) from exc


@router.head("/classes/{school}/{name}")
async def get_head(
    name: NoStarStr,
    school_authority: NonEmptyStr,
    school: NoStarStr,
    school_class: KelvinSchoolClass = Depends(get_school_class_by_name),
    session: Session = Depends(kelvin_session),
):
    await get(name, school_authority, school, school_class, session)
    return Response(status_code=status.HTTP_200_OK)


@router.post("/classes", response_model=SchoolClass, status_code=201)
async def post(
    school_authority: NonEmptyStr,
    school_class_data: SchoolClass,
    session: Session = Depends(kelvin_session),
    _is_admin: bool = Depends(check_for_authority_admin),
) -> SchoolClass:
    school_class_data = SchoolClass.with_prefix(school_class_data, school_authority)
    try:
        usernames = await member_ids_to_usernames(school_class_data.members, school_authority, session)
    except NoObject as exc:
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            detail=f"The school_class {school_class_data.name} in school {school_class_data.school} "
            f"in school_authority {school_authority} is malformed.",
        ) from exc
    school_class = KelvinSchoolClass(
        name=school_class_data.name,
        school=school_class_data.school,
        users=usernames,
        description=school_class_data.description,
        session=session,
    )
    try:
        await school_class.save()
    except InvalidRequest as exc:
        raise HTTPException(status_code=exc.status, detail=exc.reason) from exc
    # To avoid retrieving all members again, at this point we trust that the data was saved correctly
    # TODO: If performance problem is solved, use created user, not input
    return SchoolClass.without_prefix(school_class_data, school_authority)


@router.put("/classes/{school}/{name}", response_model=SchoolClass)
async def put(
    school_authority: NonEmptyStr,
    school: NonEmptyStr,
    school_class_data: SchoolClass,
    school_class: KelvinSchoolClass = Depends(get_school_class_by_name),
    session: Session = Depends(kelvin_session),
):
    if school_class_data.school != school:
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            detail="The school must not be changed after creation!",
        )
    school_class_data = SchoolClass.with_prefix(school_class_data, school_authority)
    school_class.name = school_class_data.name
    school_class.description = school_class_data.description
    try:
        usernames = await member_ids_to_usernames(school_class_data.members, school_authority, session)
    except NoObject as exc:
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            detail="Some of the users in the members field do not exist!",
        ) from exc
    school_class.users = usernames
    try:
        await school_class.save()
    except InvalidRequest as exc:
        raise HTTPException(status_code=exc.status, detail=exc.reason) from exc
    return SchoolClass.without_prefix(school_class_data, school_authority)


@router.delete("/classes/{school}/{name}", status_code=status.HTTP_204_NO_CONTENT)
async def delete(
    name: NoStarStr,
    school_authority: NonEmptyStr,
    school: NoStarStr,
    school_class: KelvinSchoolClass = Depends(get_school_class_by_name),
) -> Response:
    await school_class.delete()
    return Response(status_code=status.HTTP_204_NO_CONTENT)
