from unittest.mock import AsyncMock

import pytest
from fastapi import FastAPI
from fastapi.testclient import TestClient

import provisioning_plugin.dependencies
from id_broker_common.kelvin import kelvin_session
from provisioning_plugin.models import SchoolClass
from provisioning_plugin.plugin import router
from provisioning_plugin.routes.v1 import school_classes
from provisioning_plugin.utils import remove_prefix
from ucsschool.apis.opa import opa_instance
from ucsschool.apis.utils import auth_manager

from ..utils import Dictable, get_preconfigured_session_mock

DEFAULT_SCHOOL_AUTHORITY = "test_authority"
PREFIX = f"{DEFAULT_SCHOOL_AUTHORITY}-"
CLASS_SIZE = 5


def kelvin_school(i: int = 1) -> Dictable:
    return Dictable(
        name=f"{PREFIX}School{i}",
        display_name=f"School{i} Display",
    )


def kelvin_school_class(i: int = 1) -> Dictable:
    return Dictable(
        name=f"{i}a",
        description=f"{i}a of School1",
        school=f"{PREFIX}School1",
        users=[f"{PREFIX}student{i}" for i in range(CLASS_SIZE)],
    )


def kelvin_student(i: int = 1) -> Dictable:
    return Dictable(
        name=f"{PREFIX}student{i}",
        roles=["student"],
        firstname=f"stu{i}",
        lastname="dent",
        school_classes={f"{PREFIX}School1": ["1a"]},
        school=f"{PREFIX}School1",
        schools=[f"{PREFIX}School1"],
        ucsschool_roles=[f"student:school:{PREFIX}School1"],
        record_uid=f"student{i}",
        birthday=None,
        disabled=False,
        email=None,
        expiration_date=None,
        source_uid="TESTID",
    )


@pytest.fixture()
def broker_school():
    return Dictable(
        name="School1",
        display_name="School1 Display",
    )


@pytest.fixture()
def broker_student():
    return Dictable(
        id="student1",
        user_name="student1",
        first_name="stu1",
        last_name="dent",
        context={
            "School1": {
                "classes": ["1a"],
                "roles": ["student"],
            }
        },
    )


@pytest.fixture()
def broker_school_class():
    return Dictable(
        name="1a",
        description="1a of School1",
        school="School1",
        members=[f"student{i}" for i in range(CLASS_SIZE)],
    )


@pytest.fixture(scope="function")
def func_client(fake_opa, async_dict):
    def _client(custom_overrides):
        app = FastAPI()
        app.include_router(router)
        overrides = {
            opa_instance: fake_opa,
            auth_manager: async_dict,
            kelvin_session: get_preconfigured_session_mock(),
        }
        # we need to overwrite fast api dependencies here.
        overrides.update(custom_overrides)
        app.include_router(router)
        # let's override the dependencies, so that we don't test more than we want
        for original, override in overrides.items():
            print(f"overriding {original} with {override}")
            app.dependency_overrides[original] = override  # todo typehints?
        return TestClient(app)

    return _client


@pytest.mark.parametrize("method", ["GET", "HEAD"])
def test_user_by_user_id(func_client, method, broker_student):
    _client = func_client(
        custom_overrides={provisioning_plugin.dependencies.get_user_by_id: kelvin_student}
    )
    query = _client.get if method == "GET" else _client.head
    response = query(url=f"/v1/{DEFAULT_SCHOOL_AUTHORITY}/users/student1")
    assert response.status_code == 200
    if method == "GET":
        assert response.json() == broker_student


@pytest.mark.parametrize("method,status_code", [("POST", 201), ("PUT", 200)])
def test_put_post_user(func_client, method, status_code, broker_student):
    _client = func_client(
        custom_overrides={
            provisioning_plugin.dependencies.get_user_by_id: kelvin_student,
            kelvin_session: get_preconfigured_session_mock(
                {
                    "post.return_value": kelvin_student(),
                }
            ),
            provisioning_plugin.dependencies.check_for_authority_admin: lambda: True,
        }
    )
    url = f"/v1/{DEFAULT_SCHOOL_AUTHORITY}/users"
    query = None
    if method == "POST":
        query = _client.post
    if method == "PUT":
        query = _client.put
        url += f"/{broker_student['id']}"
    response = query(url, json=broker_student)
    assert response.status_code == status_code
    assert response.json() == broker_student


def test_delete_user(func_client, broker_student):
    _client = func_client(
        custom_overrides={
            provisioning_plugin.dependencies.get_user_by_id: kelvin_student,
            provisioning_plugin.dependencies.check_for_authority_admin: lambda: True,
        }
    )
    response = _client.delete(f"/v1/{DEFAULT_SCHOOL_AUTHORITY}/users/{broker_student['id']}")
    assert response.status_code == 204


@pytest.mark.parametrize("method", ["GET", "HEAD"])
def test_school_by_name(func_client, method, broker_school):
    _client = func_client(
        custom_overrides={provisioning_plugin.dependencies.get_school_by_name: kelvin_school}
    )
    query = _client.get if method == "GET" else _client.head
    response = query(url=f"/v1/{DEFAULT_SCHOOL_AUTHORITY}/schools/{broker_school.name}")
    assert response.status_code == 200
    if method == "GET":
        assert response.json() == broker_school


def test_post_school(func_client, broker_school):
    _client = func_client(
        custom_overrides={
            kelvin_session: get_preconfigured_session_mock(
                {
                    "post.return_value": kelvin_school(),
                }
            ),
            provisioning_plugin.dependencies.check_for_authority_admin: lambda: True,
        }
    )
    response = _client.post(f"/v1/{DEFAULT_SCHOOL_AUTHORITY}/schools", json=broker_school)
    assert response.status_code == 201
    assert response.json() == broker_school


@pytest.mark.parametrize("method", ["GET", "HEAD"])
def test_school_class_by_school_and_name(func_client, monkeypatch, method, broker_school_class):
    # 1st step: values like kelvin, but members instead of users
    exp = kelvin_school_class()
    # 2nd step: we expect {DEFAULT_SCHOOL_AUTHORITY} to be removed
    exp.members = [remove_prefix(user, PREFIX) for user in exp.users]
    monkeypatch.setattr(SchoolClass, "from_kelvin_school_class", AsyncMock(return_value=exp))
    _client = func_client(
        custom_overrides={
            provisioning_plugin.dependencies.get_school_class_by_name: kelvin_school_class,
        }
    )
    query = _client.get if method == "GET" else _client.head
    response = query(
        url=f"/v1/{DEFAULT_SCHOOL_AUTHORITY}/classes/{broker_school_class.school}/{broker_school_class.name}"
    )
    assert response.status_code == 200
    if method == "GET":
        res = response.json()
        res["members"] = sorted(res["members"])
        assert res == broker_school_class


@pytest.mark.parametrize(
    "method,status_code",
    [
        ("POST", 201),
        ("PUT", 200),
    ],
)
def test_post_school_class(func_client, monkeypatch, method, status_code, broker_school_class):
    monkeypatch.setattr(
        school_classes,
        "member_ids_to_usernames",
        AsyncMock(return_value=[kelvin_student(i) for i in range(CLASS_SIZE)]),
    )
    _client = func_client(
        custom_overrides={
            kelvin_session: get_preconfigured_session_mock(
                {
                    "post.return_value": kelvin_school_class(),
                }
            ),
            provisioning_plugin.dependencies.get_school_class_by_name: kelvin_school_class,
            provisioning_plugin.dependencies.check_for_authority_admin: lambda: True,
        }
    )
    url = f"/v1/{DEFAULT_SCHOOL_AUTHORITY}/classes"
    query = None
    if method == "POST":
        query = _client.post
    elif method == "PUT":
        query = _client.put
        url += f"/{broker_school_class.school}/{broker_school_class.name}"
    response = query(
        url,
        json=broker_school_class,
    )
    assert response.status_code == status_code
    res = response.json()
    res["members"] = sorted(res["members"])
    assert res == broker_school_class
